import greenfoot.*;

/**
 * CLASS:  GQActor (subclass of QActor, subclass of Actor)<br>
 * AUTHOR: danpost (greenfoot.org username)<br>
 * DATE: December, 26, 2014<br>
 * MODIFIED: January 8, 2014: extensive horizontal control added<br>
 * MODIFIED: January 9, 2014: image control added<br>
 * MODIFIED: May 10, 2017: moveed some intersecting code to 'touched' method and removed 'vSpeed' field
 * <br>
 * DESCRIPTION: an abstract class that can be used to give actors the effect of gravity;
 * this makes falling and jumping a breeze; methods that return the current states of the
 * vertical movement are included which will be helpful in many situations ('isOnGround',
 * 'hasJustLanded' and 'hasJustJumped'); yes, jumping is even implemented -- just call the
 * 'setJumping' method to set the speed of jump and the key that triggers jumping; there
 * is also a 'setObstacles' method that you can call to list the classes that create
 * objects that are obstacles to the actor.<br>
 * This class has been updated to add an extensive system for horizontal movement control;
 * speed, acceleration, and even drag were all taken into account as well as assigning
 * keyboard keys for left and right;
 * <br>
 * NOTE: because this class extends the QActor class, all int fields are fine-tuned values --
 * meaning that they are magnified by the zoom factor, QVAL (= 100), for accuracy.<br>
 * INSTRUCTIONS:<br>
 * <ol><li>set-up<br>
 *    To make use of the superclass, use any combination of the "set-up" methods included in
 *    the class.  There include 'setGravity', 'setObstacles', 'setJumping', 'setHorizontalMovement',
 *    and 'setDragForces'.  There is even a 'setImages' method to help control the image of the actor.
 *    An added 'setJumpImages' method was added to implement control of them also.
 * <li>use<br>
 *    Call either or both of the methods 'moveVertically' and 'moveHorizontally' (preferably in that order)
 *    from the act method of the subclass.  After the 'moveVertically' call, you can use the 'hasJustJumped',
 *    'hasJustLanded', 'isOnGround' and 'hasGroundChanged' methods to perform any action that need done when
 *    any of these actions has occurred.  In case you need to know if any of the obstacle classes objects has
 *    been touched, there is a 'touching' method that can be overridden to deal with any situation.  Getter 
 *    and setter methods are available for all pertinent fields to help with changes in these values during
 *    gameplay.
 */
public abstract class GQActor extends QActor
{
    /** ***********************************   VERTICAL MOVEMENT   ******************************* */
    
    private Class[] obstacleClasses; // list of classes that create obstacles for the actor
    private int gravity = QVAL; // the force of gravity on this actor
    private String jumpKey; // the keyboard key that triggers a jump
    private int jStrength; // the jump strength
    private boolean onGround; // state of actor on ground object or bottom edge of window
    private boolean justLanded; // state of actor just landing
    private boolean justJumped; // state of actor just jumping
    private boolean groundChanged; // state of actor just changing ground objects or bottom edge of window
    private Actor groundActor; // last Actor object this actor was found to be on
    private GreenfootImage jumpLeft, jumpRight; // images used while not on ground
    
    /**
     * controls vertical movement of this actor, which include falling and jumping; information about the movement,
     * including jumping, landing, on-ground and ground changes are recorded as well as saving the Actor last acting
     * as the ground object
     */
    public void moveVertically()
    {
        boolean previouslyOnGround = onGround; // previous 'onGround' state
        groundChanged = false; // reset flag
        justLanded = false; // reset flag
        onGround = false; // reset flag
        addForce(gravity, 90*QVAL); // acceleration due to gravity
        super.moveVertically(); // move vertically
        int dir = (int)Math.signum(getVY()); // current direction of movement
        int[] pYs = new int[] { getY(), getY() };
        for (Class clss : obstacleClasses) // for each obstacle class
        {
            for (Object obj : getIntersectingObjects(clss)) // for each intersecting obstacle
            {
                Actor touched = (Actor)obj; // cast intersector as type Actor
                touching(touched, 1); // called for subclass to act on specific obstacle types
                if (touched.getWorld() == null) continue;
                int py = touched.getY()-dir*((touched.getImage().getHeight()+getImage().getHeight())/2);
                if (pYs[(1+dir)/2]*dir > py*dir) pYs[(1+dir)/2] = py;
                if (dir > 0) // if falling
                {
                    if (!previouslyOnGround)
                    {
                        justLanded = true; // set flag}
                       if (leftImage != null) setImage(lastMovedLeft ? leftImage : rightImage); // set ground image
                    }
                    onGround = true; // set flag
                    if (groundActor != touched) groundChanged = true; // set flag if needed
                    groundActor = touched; // set reference to ground actor
                }
            }
        }
        if (pYs[(1+dir)/2] != getY()) setQLocation(getQX(), QVAL*pYs[(1+dir)/2]+99);
        int yLimit = getWorld().getHeight()-getImage().getHeight()/2; // limit actor can drop before edge of world
        if (getY() > yLimit) // if world bounds exceeded
        {
            int holdQX = getQX();
            setLocation(getX(), yLimit+1);
            setQLocation(holdQX, getQY()); // set proper distance from bottom edge
            setVY(0); // kill vertical speed
            if (!previouslyOnGround)
            {
                justLanded = true; // set flag
                if (leftImage != null && rightImage != null)
                {
                    if (lastMovedLeft) setImage(leftImage); else setImage(rightImage); // adjust image if needed
                }
            }
            onGround = true; // set flag
            if (groundActor != null) groundChanged = true; // set flag if needed
            groundActor = null; // save reference to ground actor (null for bottom edge of world)
        }
        justJumped = false; // reset flag
        if (jStrength > 0 && jumpKey != null && onGround && Greenfoot.isKeyDown(jumpKey)) // jumping
        {
            justJumped = true; // set flag
            addForce(-jStrength, 90*QVAL); // adjust vertical speed
            if (jumpRight != null && jumpLeft != null) // adjust image if available for jumping
            {
                if (lastMovedLeft) setImage(jumpLeft); else setImage(jumpRight); // set appropriate jump image
            }
        }
    }
    
    /**
     * sets the force of gravity to apply on this actor
     * 
     * @param force the amount of gravitational force to apply on this actor
     */
    public void setGravity(int force) { gravity = force; }
    
    /** returns the currently set value for the force of gravity for this actor
     * 
     * @return the value of the gravity force currently set for this actor
     */
    public int getGravity() { return gravity; }
    
    /**
     * sets the jump strength and trigger key for this actor to jump
     * 
     * @param key the keyboard key string value that triggers a jump
     * @param strength the initial jump force in the upward direction
     */
    public void setJumping(String key, int strength)
    {
        jumpKey = key;
        jStrength = (int)Math.abs(strength);
    }
    
    /**
     * sets the key used to trigger this actor to jump
     * 
     * @param key the string representation of the key to use for making this actor jump
     */
    public void setJumpKey(String key) { jumpKey = key;}
    
    /**
     * returns the currently set trigger key used for this actor to jump
     * 
     * @return the string representation of the currently set jump key
     */
    public String getJumpKey() { return jumpKey; }
    
    /**
     * sets the jump strength for this actor
     * 
     * @param strength the value to use for the jump strength of this actor
     */
    public void setJumpStrength(int strength) { jStrength = (int)Math.abs(strength); }
    
    /**
     * returns the currently set jump strength for this actor
     * 
     * @return the value of the jump strength currently set for this actor
     */
    public int getJumpStrength() { return jStrength; }
    
    /**
     * sets the images used for when this actor is not on ground
     * 
     * NOTE: the images will be scaled to be the same size as the ones used for on ground.
     * 
     * @param name the filename of the right facing image to use while this actor is not on ground (use 'null'
     * to cancel use of jump images)
     */
    public void setJumpImages(String name)
    {
        if (name == null)
        {
            jumpLeft = null;
            jumpRight = null;
            return;
        }
        jumpRight = new GreenfootImage(name);
        jumpRight.scale(rightImage.getWidth(), rightImage.getHeight());
        jumpLeft = getReverseImage(jumpRight);
    }
        
    /**
     * returns true if the actor was last found not to be falling
     * 
     * @return the last known state of this actor being on the ground
     */
    public boolean isOnGround() { return onGround; }
    
    /**
     * returns true if the actor just went from a falling state to a non-falling state
     * 
     * @return the state of this actor changing states from falling to non-falling 
     */
    public boolean hasJustLanded() { return justLanded; }
    
    /**
     * returns true if the actor just jumped
     * 
     * @return the state of this actor having initiated a jump
     */
    public boolean hasJustJumped() { return justJumped; }
    
    /**
     * returns true if the actor just moved or landed on a new ground object
     * 
     * @return the state of this actor moving or landing on a new ground object
     */
    public boolean hasGroundChanged() { return groundChanged; }
    
    /**
     * returns the ground object the actor was last found to be on; the field is initially set to 'null' and
     * will also become 'null' if the actor lands at the bottom edge of the window
     * 
     * @return the Actor object (or null) the actor was last found to be on
     */
    public Actor getGroundActor() { return groundActor; }
    

    /** **********************************   UNIVERSAL   ********************************* */
    
    private GreenfootImage leftImage, rightImage; //  the left and right facing images for the actor

    /**
     * called by subclasses, usually from the constructor, to set the list of classes that
     * create obstacles for the actor
     * 
     * @param classes a comma delimited list of classes that create obstacles for the actor
     */
    protected void setObstacles(Class... classes)
    {
        obstacleClasses = classes;
    }
    
    /**
     * called by subclasses, usually from the constructor, to set the left and right facing
     * images from the default image for the class, taking the currently set image  as the right one
     * and horizontally mirroring it for the left one
     */
    public void setImages()
    {
        rightImage = getImage();
        leftImage = getReverseImage(rightImage);
        setImage(rightImage);
    }
    
    /**
     * returns a horizontally mirrored rendition of the image given (facingLeft || facingRight)
     * 
     * @param img the image to be mirror horizontally
     * 
     * @return a horizontally mirrored rendition of the image given
     */
    public static GreenfootImage getReverseImage(GreenfootImage img)
    {
        if (img == null) return null;
        int w = img.getWidth(), h = img.getHeight();
        GreenfootImage image = new GreenfootImage(w, h);
        for (int y=0; y<h; y++) for (int x=1; x<=w; x++)
            image.setColorAt(w-x, y, img.getColorAt(x-1, y));
        return image;
    }
    
    /**
     * to be overridden by subclasses to act upon specific types of intersecting actors;
     * for example, let us say the longer you touch a wall the weaker it becomes and then breaks;
     * the following can be written in a subclass to deal with that
     * 
     * <pre>     protected void touching(Actor obstacle, int axis)
     *     {
     *         if (axis == 0 && obstacle instanceof Wall) ((Wall)obstacle).weaken();
     *         super.touched(obstacle, axis);
     *     }</pre>
     * 
     * a 'weaken' method in the Wall class can then decrease the stability of the wall and remove
     * it when needed
     */
    protected void touching(Actor touched, int axis)
    {
        if (axis == 0)
        {
            int dir = (int)Math.signum(getVX()); // current direction of movement
            int dist = (touched.getImage().getWidth()+getImage().getWidth())/2; // distance to maintain with intersecting object
            setQLocation((touched.getX()-dist*dir)*QVAL, getQY()); // set proper distance
            setVX(0); // kill horizontal speed
        }
        else
        {
            int dir = (int)Math.signum(getVY()); // current direction of movement
            int dist = (touched.getImage().getHeight()+getImage().getHeight())/2-1; // distance to maintain with intersecting object
            setQLocation(getQX(), (touched.getY()-dist*dir)*QVAL); // set proper distance
            setVY(0); // kill vertical speed
            if (!onGround) justLanded = true;
            else
            {
                justLanded = false;
                groundChanged = groundActor != touched;
            }
            onGround = true;
            groundActor = touched;
        }
    }
    
    /** *****************************   HORIZONTAL MOVEMENT   *************************** */
    
    private String leftKey, rightKey; // the keyboard keys for horizontal movement
    private int hMaxVel; // the maximum horizontal speed
    private int hAccel; // the horizontal acceleration
    private int floorDrag; // the drag when on ground
    private int airDrag; // the drag when in air
    private boolean lastMovedLeft; // flag signifying if last movement of actor was in the left direction
    
    /**
     * to be called by act method of subclasses to produce horizontal movement for the actor
     */
    public void moveHorizontally()
    {
        if (hMaxVel == 0) return; // if does not move horizontally, do not do anything
        if (onGround)
        { // on ground
            setQLocation(getQX(), getQY()-QVAL*2); // get off ground to move horizontally
            int drag = -(int)Math.signum(getVX())*floorDrag; // determine drag amount
            if (getVX()*(getVX()+drag) < 0) setVX(0); else addForce(drag, 0); // adjust or zero speed due to drag
        }
        else
        { // in air
            int drag = -(int)Math.signum(getVX())*airDrag; // determine drag amount
            if (getVX()*(getVX()+drag) < 0) setVX(0); else setVX(getVX()+drag); // adjust or zero speed due to drag
        }
        int dx = 0; // local field for user input
        if (leftKey != null && Greenfoot.isKeyDown(leftKey)) dx--;
        if (rightKey != null && Greenfoot.isKeyDown(rightKey)) dx++;
        addForce(dx*(hAccel*floorDrag), 0); // adjust speed due to user input
        if (Math.abs(getVX()) > hMaxVel) setVX((int)Math.signum(getVX())*hMaxVel); // limit speed to maximum if needed
        super.moveHorizontally(); // move actor
        if (obstacleClasses == null) return; // done if no obstacle classes given
        int dir = (int)Math.signum(getVX()); // current direction of movement
        for (Class clss : obstacleClasses) // for each obstacle class
        {
            for (Object obj : getIntersectingObjects(clss)) // for each intersecting obstacle
            {
                Actor touched = (Actor)obj; // cast intersector as type Actor
                touching(touched, 0); // called for subclass to act on specific actors or types of actors
            }
        }
        if ((dir < 0 && !lastMovedLeft) || (dir > 0 && lastMovedLeft))
            lastMovedLeft = !lastMovedLeft; // adjust flag if needed
        if (rightImage != null && !lastMovedLeft && (onGround || jumpRight == null) && getImage() != rightImage)
            setImage(rightImage); // change to right facing image if needed
        if (leftImage != null && lastMovedLeft && (onGround || jumpRight == null) && getImage() != leftImage)
            setImage(leftImage); // change to left facing image if needed
        if (jumpRight != null && jumpLeft != null)
        { // using jump images
            if (!lastMovedLeft && !onGround && getImage() != jumpRight)
                setImage(jumpRight); // change to right facing jump image if needed
            if (lastMovedLeft && !onGround && getImage() != jumpLeft)
                setImage(jumpLeft); // change to left facing jump image if needed
        }
        if (onGround) setQLocation(getQX(), getQY()+QVAL*2); // get back on ground
    }
    
    /**
     * called by subclasses, usually from the constructor to set the horizonal movement keys, max speed and acceleration
     * 
     * @param keyLeft the key to cause the actor to tend toward the left
     * @param keyRight the key to cause the actor to tend toward the right
     * @param maximumSpeed the maximum value the horizontal speed can take on
     * @param acceleration the rate of increase in speed
     */
    public void setHorizontalMovement(String keyLeft, String keyRight, int maximumSpeed, int acceleration)
    {
        leftKey = keyLeft;
        rightKey = keyRight;
        setHorizontalMovement(maximumSpeed, acceleration);
    }
    
    /**
     * sets the trigger key used to make the actor tend toward left
     * 
     * @param keyLeft the key to cause the actor to tend toward the left
     */
    public void setLeftKey(String keyLeft) { leftKey = keyLeft; }
    
    /**
     * returns the currently set trigger key used to make the actor tend toward the left
     * 
     * @return the current trigger key that makes the actor tend toward the left
     */
    public String getLeftKey() { return leftKey; }
    
    /**
     * sets the trigger key used to make the actor tend toward right
     * 
     * @param keyRight the key to cause the actor to tend toward the right
     */
    public void setRightKey(String keyRight) { rightKey = keyRight; }
    
    /**
     * returns the currently set trigger key used to make the actor tend toward the right
     * 
     * @return the current trigger key that makes the actor tend toward the right
     */
    public String getRightKey() { return rightKey; }
    
    /**
     * called by subclasses, usually from the constructor to set the max speed and acceleration of the actor;
     * may also be used to modify the behavior of the actor during gameplay
     * 
     * @param maximumSpeed the maximum value the horizontal speed can take on
     * @param acceleration the rate of increase in speed
     */
    public void setHorizontalMovement(int maximumSpeed, int acceleration)
    {
        hMaxVel = (int)Math.abs(maximumSpeed);
        hAccel = (int)Math.abs(acceleration);
    }
    
    /**
     * sets the maximum value the horizontal speed can take on
     * 
     * @param maximumSpeed the maximum value the horizontal speed can take on
     * @param acceleration the rate of increase in speed
     */
    public void setMaximumSpeed(int maximumSpeed) { hMaxVel = (int)Math.abs(maximumSpeed); }
    
    /**
     * returns the currently set maximum for the horizontal speed of this actor
     * 
     * @return the current value set for the maximum horizontal speed of this actor
     */
    public int getMaximumSpeed() { return hMaxVel; }
    
    /**
     * sets the horizontal acceleration for this actor
     * 
     * @param accel the value for the horizontal acceleration of this actor
     */
    public void setHorizontalAcceleration(int accel) { hAccel = (int)Math.abs(accel); }
    
    /**
     * returns the currently set value for the horizontal acceleration of this actor
     * 
     * @return the current horizontal acceleration value of this actor
     */
    public int getHorizontalAcceleration() { return hAccel; }
    
    /**
     * called by subclasses, usually from the constructor to set the drag values for on ground and in air
     * 
     * @param downDrag the horizontal drag value for the actor while on the ground
     * @param upDrag the horizontal drag value for the actor while not on the ground
     */
    public void setDragForces(int downDrag, int upDrag)
    {
        floorDrag = downDrag;
        airDrag = upDrag;
    }
    
    /**
     *  sets the drag value for while on ground
     *  
     * @param downDrag the horizontal drag value for the actor while on the ground
     */
    public void setGroundDrag(int downDrag) { floorDrag = downDrag; }
    
    /**
     * returns the currently set drag value for this actor while on ground
     * 
     * @return the current value of ground drag for this actor
     */
    public int getGroundDrag() { return floorDrag; }
    
    /**
     * sets the drag value for while not on ground
     * 
     * @param upDrag the horizontal drag value for the actor while not on the ground
     */
    public void setAirDrag(int upDrag) { airDrag = upDrag; }
    
    /**
     * returns the currently set drag value for this actor while not on ground
     * 
     * @return the current value of air drag for this actor
     */
    public int getAirDrag() { return airDrag; }
}